Package: imagine
Type: Package
Title: IMAGing engINEs, Tools for Application of Image Filters to Data
        Matrices
Version: 2.1.4
Date: 2026-05-13
URL: https://github.com/LuisLauM/imagine
BugReports: https://github.com/LuisLauM/imagine/issues
Authors@R: c(person("Wencheng", "Lau-Medrano", role = c("aut", "cre"), email = "luis.laum@protonmail.com"))
Maintainer: Wencheng Lau-Medrano <luis.laum@protonmail.com>
Description: Provides fast application of image filters to data matrices,
    using R and C++ algorithms.
License: GPL (>= 2)
LazyData: TRUE
Depends: R (>= 4.1.0)
Imports: cli, Rcpp, RcppArmadillo
LinkingTo: Rcpp, RcppArmadillo
Suggests: knitr, rmarkdown, quarto
VignetteBuilder: quarto, knitr
Encoding: UTF-8
RoxygenNote: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-13 16:16:05 UTC; afipdbl
Author: Wencheng Lau-Medrano [aut, cre]
Repository: CRAN
Date/Publication: 2026-05-14 14:20:18 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-14 14:35:51 UTC; unix
Archs: imagine.so.dSYM

Package: imagine
Type: Package
Title: IMAGing engINEs, Tools for Application of Image Filters to Data
        Matrices
Version: 2.1.4
Date: 2026-05-13
URL: https://github.com/LuisLauM/imagine
BugReports: https://github.com/LuisLauM/imagine/issues
Authors@R: c(person("Wencheng", "Lau-Medrano", role = c("aut", "cre"), email = "luis.laum@protonmail.com"))
Maintainer: Wencheng Lau-Medrano <luis.laum@protonmail.com>
Description: Provides fast application of image filters to data matrices,
    using R and C++ algorithms.
License: GPL (>= 2)
LazyData: TRUE
Depends: R (>= 4.1.0)
Imports: cli, Rcpp, RcppArmadillo
LinkingTo: Rcpp, RcppArmadillo
Suggests: knitr, rmarkdown, quarto
VignetteBuilder: quarto, knitr
Encoding: UTF-8
RoxygenNote: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-13 16:16:05 UTC; afipdbl
Author: Wencheng Lau-Medrano [aut, cre]
Repository: CRAN
Date/Publication: 2026-05-14 14:20:18 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-14 15:36:38 UTC; unix
Archs: imagine.so.dSYM

# Generated by roxygen2: do not edit by hand

export(agenbagFilters)
export(contextualMF)
export(convolution2D)
export(convolutionMedian)
export(convolutionQuantile)
export(meanFilter)
export(medianFilter)
export(quantileFilter)
importFrom(Rcpp,evalCpp)
importFrom(RcppArmadillo,armadillo_version)
importFrom(cli,cli_abort)
importFrom(cli,cli_warn)
useDynLib(imagine, .registration = TRUE)

## -----------------------------------------------------------------------------
#| label: setup
library(imagine)


## -----------------------------------------------------------------------------
#| eval: false
# install.packages("imagine")


## -----------------------------------------------------------------------------
#| eval: false

# # Build kernels
# # Kernel 1: For bottom edge recognition
# kernel1 <- matrix(
#   data = c(-1, -2, -1,
#             0,  0,  0,
#             1,  2,  1),
#   nrow = 3
# )
# 
# # Kernel 2: Diagonal weighting
# kernel2 <- matrix(
#   data = c(-2, 0, 0,
#             0, 1, 0,
#             0, 0, 2),
#   nrow = 3
# )
# 
# # Apply filters
# convolutionExample  <- convolution2D(X = wbImage, kernel = kernel1)
# convQuantileExample <- convolutionQuantile(X = wbImage, kernel = kernel2, probs = 0.1)


## -----------------------------------------------------------------------------
#| message: false
#| label: fig-2d-quantile-conv
#| fig-width: 5.33
#| fig-height: 8.5
#| fig-cap: "2D vs 2D quantile convolutions"
#| fig-pos: "h"
#| results: hide
#| echo: false

# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Build kernels
# Kernel 1: For bottom edge recognition
kernel1 <- matrix(
  data = c(-1, -2, -1,
            0,  0,  0,
            1,  2,  1), 
  nrow = 3
)

# Kernel 2: Diagonal weighting 
kernel2 <- matrix(
  data = c(-2, 0, 0,
            0, 1, 0,
            0, 0, 2), 
  nrow = 3
)

# Apply filters
convolutionExample  <- convolution2D(X = myMatrix, kernel = kernel1)
convQuantileExample <- convolutionQuantile(
  X = myMatrix, 
  kernel = kernel2, 
  probs = 0.8
)

# Make plots
par(
  mar = c(0, 0.5, 0, 0.5), 
  oma = c(0, 0, 2, 0), 
  mfrow = c(3, 1)
)

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(convolutionExample, col = cols, axes = FALSE)
mtext(
  text = "2D convolution", 
  side = 3, 
  line = -1.5, 
  col = "white", 
  font = 2, 
  adj = 0.99
)

image(convQuantileExample, col = cols, axes = FALSE)
mtext(
  text = "2D quantile convolution", 
  side = 3, 
  line = -1.5, 
  col = "black", 
  font = 2, 
  adj = 0.99
)


## -----------------------------------------------------------------------------
#| eval: false

# # Add some noise (NA) to the image (matrix)
# set.seed(7)
# naIndex <- sample(
#   x = seq(prod(dim(myMatrix))),
#   size = as.integer(0.4*prod(dim(myMatrix))),
#   replace = FALSE
# )
# myMatrix[naIndex] <- NA
# 
# # Build kernel
# radius <- 3
# 
# # Apply filters
# meanfilterExample     <- meanFilter(X = myMatrix, radius = radius)
# quantilefilterExample <- quantileFilter(X = myMatrix, radius = radius, probs = 0.1)
# medianfilterExample   <- medianFilter(X = myMatrix, radius = radius)


## -----------------------------------------------------------------------------
#| message: false
#| label: fig-basic-filter-comparison
#| fig-width: 7.5
#| fig-height: 5
#| fig-cap: "Basic filters comparison"
#| fig-pos: "h"
#| results: hide
#| echo: false

# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Add some noise (NA) to the image (matrix)
set.seed(7)
naIndex <- sample(
  x = seq(prod(dim(myMatrix))), 
  size = as.integer(0.4*prod(dim(myMatrix)))
)
myMatrix[naIndex] <- NA

# Apply filters
meanfilterExample   <- meanFilter(X = myMatrix, radius = 3)
medianfilterExample <- medianFilter(X = myMatrix, radius = 3)

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(myMatrix, col = cols, axes = FALSE)
mtext(
  text = "Original with noise (NA)", 
  side = 3, 
  line = -1.5, 
  font = 2, 
  adj = 0.99
)

# meanfilterExample[meanfilterExample < 0] <- 0
image(meanfilterExample, col = cols, axes = FALSE)
mtext(
  text = "Mean filter", 
  side = 3, 
  line = -1.5, 
  font = 2, 
  adj = 0.99
)

# medianfilterExample[medianfilterExample < 0] <- 0
image(medianfilterExample, col = cols, axes = FALSE)
mtext(
  text = "2D median filter", 
  side = 3, 
  line = -1.5, 
  font = 2, 
  adj = 0.99
)


## -----------------------------------------------------------------------------
#| eval: false

# medianFilter(X = wbImage, radius = 5, times = 50)


## -----------------------------------------------------------------------------
#| message: false
#| label: fig-recursive-filtering
#| fig-width: 7.5
#| fig-height: 5
#| fig-cap: "Filters with several time settings"
#| fig-pos: "h"
#| results: hide
#| echo: false

times <- c(1, 5, 15)

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Apply filters
median_times1 <- medianFilter(X = wbImage, radius = 5, times = times[1])
median_times2 <- medianFilter(X = wbImage, radius = 5, times = times[2])
median_times3 <- medianFilter(X = wbImage, radius = 5, times = times[3])

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(median_times1, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[1]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times2, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[2]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times3, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[3]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(imagine)

## ----eval=FALSE---------------------------------------------------------------
# install.packages("imagine")

## ----eval=FALSE---------------------------------------------------------------
# # Build kernels
# # Kernel 1: For bottom edge recognition
# kernel1 <- matrix(c(-1, -2, -1,
#                      0,  0,  0,
#                      1,  2,  1),
#                   nrow = 3)
# 
# # Kernel 2: Diagonal weighting
# kernel2 <- matrix(c(-2, 0, 0,
#                      0, 1, 0,
#                      0, 0, 2),
#                   nrow = 3)
# 
# # Apply filters
# convolutionExample  <- convolution2D(X = wbImage, kernel = kernel1)
# convQuantileExample <- convolutionQuantile(X = wbImage, kernel = kernel2, probs = 0.1)

## ----message=FALSE, fig.height=6, fig.width=5.33, fig.cap = "Figure 1: 2D vs 2D quantile convolutions", results='hide', fig.pos="h", echo=FALSE----

# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 1, end = 0)

# Build kernels
# Kernel 1: For bottom edge recognition
kernel1 <- matrix(c(-1, -2, -1,
                     0,  0,  0,
                     1,  2,  1), 
                  nrow = 3)

# Kernel 2: Diagonal weighting
kernel2 <- matrix(c(-2, 0, 0,
                     0, 1, 0,
                     0, 0, 2), 
                  nrow = 3)

# Apply filters
convolutionExample  <- convolution2D(X = myMatrix, kernel = kernel1)
convQuantileExample <- convolutionQuantile(X = myMatrix, kernel = kernel2, probs = 0.8)

# Make plots
par(mar = c(0, 0.5, 0, 0.5), oma = c(0, 0, 2, 0), mfrow = c(2, 1))

image(convolutionExample, col = cols, axes = FALSE)
mtext(text = "2D convolution", side = 1, line = -1.5, col = "white", font = 2, adj = 0.99)

image(convQuantileExample, col = cols, axes = FALSE)
mtext(text = "2D quantile convolution", side = 1, line = -1.5, col = "black", font = 2, adj = 0.99)

## ----eval=FALSE---------------------------------------------------------------
# # Add some noise (NA) to the image (matrix)
# set.seed(7)
# naIndex <- sample(x       = seq(prod(dim(myMatrix))),
#                   size    = as.integer(0.4*prod(dim(myMatrix))),
#                   replace = FALSE)
# myMatrix[naIndex] <- NA
# 
# # Build kernel
# radius <- 3
# 
# # Apply filters
# meanfilterExample     <- meanFilter(X = myMatrix, radius = radius)
# quantilefilterExample <- quantileFilter(X = myMatrix, radius = radius, probs = 0.1)
# medianfilterExample   <- medianFilter(X = myMatrix, radius = radius)

## ----message=FALSE, fig.height=5, fig.width=7.5, fig.cap = "Figure 2: Basic filters comparison", results='hide', fig.pos="h", echo=FALSE----
# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Add some noise (NA) to the image (matrix)
set.seed(7)
naIndex <- sample(x = seq(prod(dim(myMatrix))), size = as.integer(0.4*prod(dim(myMatrix))))
myMatrix[naIndex] <- NA

# Apply filters
meanfilterExample   <- meanFilter(X = myMatrix, radius = 3)
medianfilterExample <- medianFilter(X = myMatrix, radius = 3)

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(myMatrix, col = cols, axes = FALSE)
mtext(text = "Original with noise (NA)", side = 3, line = -1.5, font = 2, adj = 0.99)

# meanfilterExample[meanfilterExample < 0] <- 0
image(meanfilterExample, col = cols, axes = FALSE)
mtext(text = "Mean filter", side = 3, line = -1.5, font = 2, adj = 0.99)

# medianfilterExample[medianfilterExample < 0] <- 0
image(medianfilterExample, col = cols, axes = FALSE)
mtext(text = "2D median filter", side = 3, line = -1.5, font = 2, adj = 0.99)

## ----eval=FALSE---------------------------------------------------------------
# medianFilter(X = wbImage, radius = 5, times = 50)

## ----message=FALSE, fig.height=5, fig.width=7.5, fig.cap = "Figure 3: Filters with several time settings", results='hide', fig.pos="h", echo=FALSE----

times <- c(1, 5, 15)

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Apply filters
median_times1 <- medianFilter(X = wbImage, radius = 5, times = times[1])
median_times2 <- medianFilter(X = wbImage, radius = 5, times = times[2])
median_times3 <- medianFilter(X = wbImage, radius = 5, times = times[3])

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(median_times1, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[1]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times2, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[2]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times3, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[3]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

checkArgs <- function(imagineArgs, type){

  switch(
    type,
    convolution2D       = checkArgs_convolution2D(imagineArgs),
    convolutionQuantile = checkArgs_convolutionQuantile(imagineArgs),
    meanFilter          = checkArgs_meanFilter(imagineArgs),
    quantileFilter      = checkArgs_quantileFilter(imagineArgs),
    contextualMF        = checkArgs_contextualMF(imagineArgs),
    agenbagFilters      = checkArgs_agenbagFilters(imagineArgs),
    "Invalid value for 'type'."
  )
}


checkArgs_convolution2D <- function(allArgs){

  X <- allArgs$X
  kernel <- allArgs$kernel
  times <- allArgs$times

  if(!is.matrix(X) || !is.numeric(X)){
    cli_abort(message = "{.code X} must be a numeric matrix.")
  }

  if(!is.matrix(kernel) || !is.numeric(kernel)){
    cli_abort(message = "{.code kernel} must be a numeric matrix.")
  }

  if((nrow(kernel) >= nrow(X)) || ncol(kernel) >= ncol(X)){
    cli_abort(message = "Dimensions of {.code kernel} matrix must be less than the {.code X}.")
  }

  if(length(times) != 1 || !is.numeric(times)){
    cli_abort("{.code time} must be a numeric vector with length 1.")
  }

  if(!isTRUE(all.equal(times, as.integer(times)))){
    cli_warn(
      message = c(
        "{.code times} must be an integer number.",
        "i" = "It will be coerced as integer."
      )
    )

    allArgs$times <- as.integer(allArgs$times)
  }

  allArgs
}

checkArgs_convolutionQuantile <- function(allArgs){

  X <- allArgs$X
  kernel <- allArgs$kernel
  probs <- allArgs$probs
  times <- allArgs$times

  if(!is.matrix(X) || !is.numeric(X)){
    cli_abort(message = "{.code X} must be a numeric matrix.")
  }

  if(any(is.na(kernel))){
    cli_abort("{.code kernel} must not have NA.")
  }

  if(!is.matrix(kernel) || !is.numeric(kernel)){
    cli_abort(message = "{.code kernel} must be a numeric matrix.")
  }

  if((nrow(kernel) >= nrow(X)) || ncol(kernel) >= ncol(X)){
    cli_abort("Dimensions of {.code kernel} matrix must be less than the {.code X}.")
  }

  if(!is.numeric(probs) || length(probs) != 1 || probs < 0 || probs > 1){
    cli_abort("{.code probs} must be a numeric between 0 and 1 with length = 1.")
  }

  if(length(times) != 1 || !is.numeric(times)){
    cli_abort("{.code time} must be a numeric vector with length 1.")
  }

  if(!isTRUE(all.equal(times, as.integer(times)))){
    cli_warn(
      message = c(
        "{.code times} must be an integer number",
        "i" = "It will be coerced as integer."
      )
    )
    allArgs$times <- as.integer(allArgs$times)
  }

  return(allArgs)
}

checkArgs_meanFilter <- function(allArgs){

  X <- allArgs$X
  radius <- allArgs$radius
  times <- allArgs$times

  if(!is.matrix(X) || !is.numeric(X)){
    cli_abort(message = "{.code X} must be a numeric matrix.")
  }

  if(!is.atomic(radius) || !is.numeric(radius)){
    cli_abort("{.code radius} must be a numeric vector of length 1 or 2.")
  }

  if(radius[1] > nrow(X) || radius[2] > ncol(X)){
    cli_abort("{.code radius} must be less than the dimensions (number of row and columns) of 'X'.")
  }

  if(!isTRUE(all.equal(radius, as.integer(radius)))){
    cli_warn("{.code radius} must be an integer number. It will be coerced as integer.")
    allArgs$radius <- as.integer(allArgs$radius)
  }

  if(length(times) != 1 || !is.numeric(times)){
    cli_abort("{.code time} must be a numeric vector with length 1.")
  }

  if(!isTRUE(all.equal(times, as.integer(times)))){
    cli_warn("{.code times} must be an integer number. It will be coerced as integer.")
    allArgs$times <- as.integer(allArgs$times)
  }

  times <- as.integer(times)

  return(allArgs)
}

checkArgs_quantileFilter <- function(allArgs){

  X <- allArgs$X
  radius <- allArgs$radius
  probs <- allArgs$probs
  times <- allArgs$times

  if(!is.matrix(X) || !is.numeric(X)){
    cli_abort(message = "{.code X} must be a numeric matrix.")
  }

  if(!is.atomic(radius) || !is.numeric(radius)){
    cli_abort("{.code radius} must be a numeric vector of length 1 or 2.")
  }

  if(!isTRUE(all.equal(radius, as.integer(radius)))){
    cli_warn("{.code radius} must be an integer vector. It will be coerced as integer.")
    allArgs$radius <- as.integer(allArgs$radius)
  }

  if(any(radius < 1) || radius[1] > nrow(X) || radius[2] > ncol(X)){
    cli_abort("Some value in {.code radius} is out of acceptable bounds in regards to 'X'.")
  }

  if(!is.numeric(probs) || length(probs) != 1 || probs < 0 || probs > 1){
    cli_abort("{.code probs} must be a numeric between 0 and 1 with length = 1.")
  }

  if(length(times) != 1 || !is.numeric(times)){
    cli_abort("{.code time} must be a numeric vector with length 1.")
  }

  if(!isTRUE(all.equal(times, as.integer(times)))){
    cli_warn("{.code times} must be an integer number. It will be coerced as integer.")
    allArgs$times <- as.integer(allArgs$times)
  }

  return(allArgs)
}

checkArgs_contextualMF <- function(allArgs){
  X <- allArgs$X

  if(!is.matrix(X) || !is.numeric(X)){
    cli_abort(message = "{.code X} must be a numeric matrix.")
  }

  return(allArgs)
}

checkArgs_agenbagFilters <- function(allArgs){

  X <- allArgs$X
  algorithm <- as.numeric(allArgs$algorithm[1])

  if(!is.matrix(X) || !is.numeric(X)){
    cli_abort(message = "{.code X} must be a numeric matrix.")
  }

  if(length(algorithm) != 1 || is.na(algorithm) || !algorithm %in% 1:2){
    cli_abort("{.code algorithm} argument only accepts next numeric values: 1 or 2.")
  }

  return(allArgs)
}

#' @title IMAGing engINE, Tools for application of image filters to data matrices
#' @description Provides fast application of image filters to data matrices,
#' using R and C++ algorithms.
#' @details This package uses C++ algorithms called 'engines'. More details are
#' shown in the vignette.
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom cli cli_abort
#' @importFrom cli cli_warn
## usethis namespace: end
NULL

#' @title Data matrix to be used as example image.
#' @description \code{matrix} object containing numeric data to plot a image.
#' The photo was taken by the author on 2016.
#' @docType data
#' @usage wbImage
#' @format A \code{matrix} with dimensions 1280x720.
"wbImage"


#' @title Make convolution calculations from numeric matrix
#'
#' @rdname convolutions
#'
#' @param X A numeric \code{matrix} object used for apply filters.
#' @param kernel A little matrix used as mask for each cell of \code{X}.
#' @param probs \code{numeric} vector of probabilities with values in [0,1].
#' @param times How many times do you want to apply the filter?
#' @param normalize \code{logical} indicating if results will (or not) be
#' normalized. See details.
#' @param na_only \code{logical}, Do you want to apply the filter ONLY in cells
#' with NA?
#'
#' @description This function takes a \code{matrix} object, and for each cell
#' multiplies its neighborhood by the \code{kernel}. Finally, it returns for
#' each cell the mean of the kernel-weighted sum.
#'
#' @return \code{convolution2D} returns a \code{matrix} object with the same
#' dimensions of \code{X}.
#'
#' @details
#' Convolution is a mathematical operation that combines two arrays of numbers
#' to produce an array of the same structure. The output will consist of only
#' valid values, meaning those where both arrays have non-missing data.
#' Consequently, any missing values (NAs) in the input matrix will propagate
#' outwards to the extent of the convolution kernel.
#'
#' Through normalization, the output of each convolution window is scaled by
#' dividing it by the sum of the absolute values of the kernel
#' (\code{sum(abs(as.numeric(kernel)))}, disabled by default).
#'
#' \code{na_only} performs two actions at once: (1) it applies the filter only in
#' the positions where the original value is \code{NA} and (2) for the rest of
#' the cells, it is replaced with the value of the original matrix.
#'
#'
#' @export
#'
#' @examples
#' # Generate example matrix
#' nRows <- 50
#' nCols <- 100
#'
#' myMatrix <- matrix(runif(nRows*nCols, 0, 100), nrow = nRows, ncol = nCols)
#' kernel <- diag(3)
#'
#' # Make convolution
#' myOutput1 <- convolution2D(myMatrix, kernel)
#' myOutput2 <- convolutionQuantile(myMatrix, kernel, probs = 0.7)
#'
#' # Plot results
#' par(mfrow = c(2, 2))
#' image(myMatrix, zlim = c(0, 100))
#' image(myOutput1, zlim = c(0, 100))
#' image(myOutput2, zlim = c(0, 100))
convolution2D <- function(X, kernel, times = 1, normalize = FALSE, na_only = FALSE){

  armadillo_version(FALSE)

  # Check and validation of arguments
  checkedArgs <- list(X = X, kernel = kernel, times = times)
  checkedArgs <- checkArgs(imagineArgs = checkedArgs, type = "convolution2D")

  # Apply filters
  output <- checkedArgs$X
  for(i in seq(checkedArgs$times)){
    gc(reset = TRUE)

    output <- with(
      checkedArgs,
      engine1_2dConv(
        data = output,
        kernel = kernel,
        na_only = isTRUE(na_only)
      )
    )

    if(normalize){
      output <- output/sum(abs(as.numeric(kernel)), na.rm = TRUE)
    }
  }

  output
}

#' @rdname convolutions
#' @return \code{convolutionQuantile} uses the kernel but, for each cell, it
#' returns the position of quantile 'probs' (value between 0 and 1).
#' @export
convolutionQuantile <- function(X, kernel, probs, times = 1, normalize = FALSE, na_only = FALSE){

  # Check and validation of arguments
  checkedArgs <- list(X = X, kernel = kernel, probs = probs, times = times)
  checkedArgs <- checkArgs(imagineArgs = checkedArgs, type = "convolutionQuantile")

  # Apply filters
  output <- checkedArgs$X
  for(i in seq(checkedArgs$times)){
    gc(reset = TRUE)

    output <- with(
      checkedArgs,
      engine2_convWithQuantiles(
        data = output,
        kernel = kernel,
        probs = probs,
        na_only = isTRUE(na_only)
      )
    )

    if(normalize){
      output <- output/sum(abs(as.numeric(kernel)), na.rm = TRUE)
    }
  }

  output
}


#' @rdname convolutions
#' @return \code{convolutionMedian} is a wrapper of \code{convolutionQuantile}
#' with probs = 0.5.
#' @export
convolutionMedian <- function(X, kernel, times = 1, na_only = FALSE){

  convolutionQuantile(
    X = X,
    kernel = kernel,
    probs = 0.5,
    times = times,
    na_only = isTRUE(na_only)
  )
}


#' @title Make a 2D filter calculations from numeric matrix
#'
#' @rdname basic2DFilter
#'
#' @param X A numeric \code{matrix} object used for apply filters.
#' @param radius Size of squared or rectangular kernel to apply median. See
#' Details.
#' @param probs \code{numeric} vector of probabilities with values in [0,1].
#' @param times How many times do you want to apply the filter?
#' @param na_only \code{logical}, Do you want to apply the filter ONLY in cells
#' with NA?
#'
#' @description This functions take a \code{matrix} object, and for each cell
#' calculate mean, median or certain quantile around a squared/rectangular
#' neighborhood.
#'
#' @return A \code{matrix} object with the same dimensions of \code{X}.
#'
#' @details \code{radius} must be defined as a 2-length numeric vector
#' specifying the number of rows and columns of the window which will be used to
#' make calculations. If the length of radius is 1, the window will be a square.
#'
#' Functions use C++ algorithms for running some statistical calculations. The
#' mean is far obvious, however, there are several ways to perform quantiles.
#' \code{quantileFilter} function uses
#' \href{https://arma.sourceforge.net/docs.html#quantile}{arma::quantile}: a
#' RcppArmadillo function, which is equivalent to use R \link[stats]{quantile}
#' funtion with \code{type = 5}.
#'
#' \code{medianFilter} is a wraper of \code{quantileFilter}, so the same
#' observations are applied to it.
#'
#' \code{na_only} performs two actions at once: (1) it applies the filter only in
#' the positions where the original value is \code{NA} and (2) for the rest of
#' the cells, it is replaced with the value of the original matrix.
#'
#' @export
#'
#' @examples
#' # Generate example matrix
#' nRows <- 50
#' nCols <- 100
#'
#' myMatrix <- matrix(runif(nRows*nCols, 0, 100), nrow = nRows, ncol = nCols)
#' radius <- 3
#'
#' # Make convolution
#' myOutput1 <- meanFilter(X = myMatrix, radius = radius)
#' myOutput2 <- quantileFilter(X = myMatrix, radius = radius, probs = 0.1)
#' myOutput3 <- medianFilter(X = myMatrix, radius = radius)
#'
#' # Plot results
#' par(mfrow = c(2, 2))
#' image(myMatrix, zlim = c(0, 100), title = "Original")
#' image(myOutput1, zlim = c(0, 100), title = "meanFilter")
#' image(myOutput2, zlim = c(0, 100), title = "quantileFilter")
#' image(myOutput3, zlim = c(0, 100), title = "medianFilter")
meanFilter <- function(X, radius, times = 1, na_only = FALSE){

  # Check and validation of arguments
  checkedArgs <- list(X = X,
                      radius = rep(x = radius, length.out = 2),
                      times = times)
  checkedArgs <- checkArgs(imagineArgs = checkedArgs, type = "meanFilter")

  # Apply filters
  output <- checkedArgs$X
  for(i in seq(checkedArgs$times)){
    gc(reset = TRUE)

    output <- with(
      checkedArgs,
      engine3_meanFilter(
        data = output,
        radius = radius,
        na_only = isTRUE(na_only)
      )
    )
  }

  output
}


#' @rdname basic2DFilter
#' @return \code{quantileFilter} don't use a kernel but, for each cell, it
#' returns the position of quantile 'probs' (value between 0 and 1).
#' @export
quantileFilter <- function(X, radius, probs, times = 1, na_only = FALSE){

  # Check and validation of arguments
  checkedArgs <- list(X = X,
                      radius = rep(x = radius, length.out = 2),
                      probs = probs,
                      times = times,
                      na = NA)
  checkedArgs <- checkArgs(imagineArgs = checkedArgs, type = "quantileFilter")

  # Apply filters
  output <- checkedArgs$X
  for(i in seq(checkedArgs$times)){
    gc(reset = TRUE)

    output <- with(
      checkedArgs,
      engine4_quantileFilter(
        data = output,
        radius = radius,
        probs = probs,
        na_only = isTRUE(na_only)
      )
    )
  }

  output
}


#' @rdname basic2DFilter
#' @return \code{medianFilter} is a wrapper of \code{quantileFilter} with
#' \code{probs = 0.5}.
#' @export
medianFilter <- function(X, radius, times = 1, na_only = FALSE){

  quantileFilter(
    X = X,
    radius = radius,
    probs = 0.5,
    times = times,
    na_only = na_only
  )
}


#' @title Performs Contextual Median Filter
#'
#' @param X A numeric \code{matrix} object used for apply filters.
#'
#' @description This function implements the Contextual Median Filter (CMF)
#' algorithm, which was first described by Belkin & O'Reilly (2009), following
#' the pseudocode provided in their paper.
#'
#' @details
#' Following the definition of CMF, since \strong{imagine} v.2.0.0, \code{times}
#' argument will not be available anymore.
#'
#' \strong{imagine} offers the CMF algorithm but for the using to find out
#' oceanographic fronts, it is recommended to see and use the functions of the
#' \href{https://CRAN.R-project.org/package=grec}{\strong{grec}} package.
#'
#' @references Belkin, I. M., & O'Reilly, J. E. (2009). An algorithm for oceanic
#' front detection in chlorophyll and SST satellite imagery. Journal of Marine
#' Systems, 78(3), 319-326 (\doi{http://dx.doi.org/10.1016/j.jmarsys.2008.11.018}).
#'
#' @export
#'
#' @return \code{contextualMF} returns a \code{matrix} object with the same
#' dimensions of \code{X}.
#'
#' @examples
#' data(wbImage)
#'
#' cmfOut <- contextualMF(X = wbImage)
#'
#' # image(cmfOut)
contextualMF <- function(X){

  # Check and validation of arguments
  checkedArgs <- list(X = X)
  checkedArgs <- checkArgs(imagineArgs = checkedArgs, type = "contextualMF")

  # Apply filters
  with(checkedArgs, engine5_CMF(data = X))
}


#' @title Performs algorithms from Agenbag et al. (2003)
#'
#' @param X A numeric \code{matrix} used as main input.
#' @param algorithm \code{integer} specifying the type of method that will be
#' used. See Details.
#' @param ... Not used.
#'
#' @description This function performs two (gradient) calculation approaches for
#' SST, as outlined in the paper by Agenbag et al. (2003).
#'
#' @details
#' Section 2.2.4 of the paper by Agenbag et al. (2003) introduces the following
#' two methods:
#' \describe{
#'  \item{\strong{Method 1: }}{Based on the equation
#'  \deqn{Y_{i,j}=\sqrt{(X_{i+1,j}-X_{i-1,j})^2 +(X_{i,j+1}-X_{i,j-1})^2}}}
#'  where \eqn{Y_{i,j}} represents the output value for each \eqn{X_{i,j}} pixel value
#'  of a given \eqn{X} matrix.
#'  \item{\strong{Method 2: }}{the standard deviation in a 3x3 pixel area centered on
#'  position \eqn{(i,j)}.}
#' }
#'
#' As outlined in the original study, this method conducts searches within a
#' 1-pixel vicinity of each point. For method 1, it only returns a value for
#' points where none of the four involved values are NA. Conversely, for method
#' 2, the standard deviation calculation is performed only for points where at
#' least 3 non-NA values are found in the 3x3 neighborhood.
#'
#'
#' @references Agenbag, J.J., A.J. Richardson, H. Demarcq, P. Freon, S. Weeks,
#' and F.A. Shillington. "Estimating Environmental Preferences of South African
#' Pelagic Fish Species Using Catch Size- and Remote Sensing Data". Progress in
#' Oceanography 59, No 2-3 (October 2003): 275-300.
#' (\doi{https://doi.org/10.1016/j.pocean.2003.07.004}).
#'
#' @return \code{agenbagFilters} returns a \code{matrix} object with the same
#' dimensions of \code{X}.
#'
#' @export
#'
#' @examples
#' data(wbImage)
#'
#' # Agenbag, method 1
#' agenbag1 <- agenbagFilters(X = wbImage, algorithm = 1)
#'
#' # Agenbag, method 2
#' agenbag2 <- agenbagFilters(X = wbImage, algorithm = 2)
#'
#' # Plotting results
#' par(mfrow = c(3, 1), mar = rep(0, 4))
#'
#' # Original
#' image(wbImage, axes = FALSE, col = gray.colors(n = 1e3))
#'
#' # Calculated
#' cols <- hcl.colors(n = 1e3, palette = "YlOrRd", rev = TRUE)
#' image(agenbag1, axes = FALSE, col = cols)
#' image(agenbag2, axes = FALSE, col = cols)
agenbagFilters <- function(X, algorithm = c(1, 2), ...){

  # Check and validation of arguments
  checkedArgs <- list(X = X,
                      algorithm = algorithm)
  checkedArgs <- checkArgs(imagineArgs = checkedArgs, type = "agenbagFilters")

  switch(
    checkedArgs$algorithm,
    "1" = engine6_agenbag1(data = checkedArgs$X),
    "2" = engine7_agenbag2(data = checkedArgs$X, ...)
  )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @importFrom Rcpp evalCpp
#' @useDynLib imagine, .registration = TRUE
NULL

engine1_2dConv <- function(data, kernel, na_only) {
    .Call(`_imagine_engine1_2dConv`, data, kernel, na_only)
}

#' @importFrom Rcpp evalCpp
#' @importFrom RcppArmadillo armadillo_version
#' @useDynLib imagine, .registration = TRUE
NULL

engine6_agenbag1 <- function(data) {
    .Call(`_imagine_engine6_agenbag1`, data)
}

engine7_agenbag2 <- function(data, threshold = 3L) {
    .Call(`_imagine_engine7_agenbag2`, data, threshold)
}

#' @importFrom Rcpp evalCpp
#' @useDynLib imagine, .registration = TRUE
NULL

engine5_CMF <- function(data, i_size = 3L, o_size = 5L) {
    .Call(`_imagine_engine5_CMF`, data, i_size, o_size)
}

#' @importFrom Rcpp evalCpp
#' @importFrom RcppArmadillo armadillo_version
#' @useDynLib imagine, .registration = TRUE
NULL

engine2_convWithQuantiles <- function(data, kernel, probs, na_only) {
    .Call(`_imagine_engine2_convWithQuantiles`, data, kernel, probs, na_only)
}

#' @importFrom Rcpp evalCpp
#' @useDynLib imagine, .registration = TRUE
NULL

engine3_meanFilter <- function(data, radius, na_only) {
    .Call(`_imagine_engine3_meanFilter`, data, radius, na_only)
}

#' @importFrom Rcpp evalCpp
#' @importFrom RcppArmadillo armadillo_version
#' @useDynLib imagine, .registration = TRUE
NULL

engine4_quantileFilter <- function(data, radius, probs, na_only) {
    .Call(`_imagine_engine4_quantileFilter`, data, radius, probs, na_only)
}

---
title: "imagine-intro"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{imagine-intro}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(imagine)
```

# imagine

**IMAGing engINE, Tools for Application of Image Filters to Data Matrices**

The **imagine** package streamlines the process of applying image-filtering algorithms to numeric data matrices. It employs efficient median-filter and 2D-convolution algorithms implemented in Rcpp (C++), enabling rapid processing of large datasets.


## Installation

For installing `imagine`, as follows:

```{r, eval=FALSE}
install.packages("imagine")
```

## Engines
The **imagine** package employs C++-based algorithms, designated as 'engines,' crafted with [**Rcpp**](https://cran.r-project.org/package=Rcpp) and [**RcppArmadillo**](https://cran.r-project.org/package=RcppArmadillo), to expedite the application of image filters. These engines significantly enhance the performance of filtering operations, ensuring efficient processing of large datasets. As of version 2.1.0, **imagine** incorporates the following engines:

* Engine 1: The basic 2D-convolution operation that involves multiplying a kernel by the neighborhood of each cell and summing the results.
* Engine 2: it performs the same convolution operation as Engine 1 but returns the position specified by the `probs` parameter.
* Engine 3: It uses the `radius` argument to extract the values of squared neighborhood ($radius \times radius$) and calculates the mean.
* Engine 4: It uses the `radius` argument to extract the values of squared neighborhood ($radius \times radius$) and returns the position indicated by the parameter `probs`
* Engine 5: Contextual Median Filter described in Belkin and O'Reilly (2009) paper <http://dx.doi.org/10.1016/j.jmarsys.2008.11.018>.
* Engines 6 and 7: Gradient filters proposed by Agenbag et al. (2003) paper <https://doi.org/10.1016/j.pocean.2003.07.004>.

Since version 2.0.0, `radius` could accept 2 values to define the number of rows and columns respectively of the window.

## Main functions

There are 5 main functions and 2 wrappers:

### Convolution functions
```{r, eval=FALSE}
# Build kernels
# Kernel 1: For bottom edge recognition
kernel1 <- matrix(c(-1, -2, -1,
                     0,  0,  0,
                     1,  2,  1), 
                  nrow = 3)

# Kernel 2: Diagonal weighting
kernel2 <- matrix(c(-2, 0, 0,
                     0, 1, 0,
                     0, 0, 2), 
                  nrow = 3)

# Apply filters
convolutionExample  <- convolution2D(X = wbImage, kernel = kernel1)
convQuantileExample <- convolutionQuantile(X = wbImage, kernel = kernel2, probs = 0.1)
```

In order to compare results, we will plot both data (original and filtered) using `image` function, as shows in figures 1 and 2.

**Original vs filtered outputs**

```{r, message=FALSE, fig.height=6, fig.width=5.33, fig.cap = "Figure 1: 2D vs 2D quantile convolutions", results='hide', fig.pos="h", echo=FALSE}

# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 1, end = 0)

# Build kernels
# Kernel 1: For bottom edge recognition
kernel1 <- matrix(c(-1, -2, -1,
                     0,  0,  0,
                     1,  2,  1), 
                  nrow = 3)

# Kernel 2: Diagonal weighting
kernel2 <- matrix(c(-2, 0, 0,
                     0, 1, 0,
                     0, 0, 2), 
                  nrow = 3)

# Apply filters
convolutionExample  <- convolution2D(X = myMatrix, kernel = kernel1)
convQuantileExample <- convolutionQuantile(X = myMatrix, kernel = kernel2, probs = 0.8)

# Make plots
par(mar = c(0, 0.5, 0, 0.5), oma = c(0, 0, 2, 0), mfrow = c(2, 1))

image(convolutionExample, col = cols, axes = FALSE)
mtext(text = "2D convolution", side = 1, line = -1.5, col = "white", font = 2, adj = 0.99)

image(convQuantileExample, col = cols, axes = FALSE)
mtext(text = "2D quantile convolution", side = 1, line = -1.5, col = "black", font = 2, adj = 0.99)
```


### Median-filter asociated functions
```{r, eval=FALSE}
# Add some noise (NA) to the image (matrix)
set.seed(7)
naIndex <- sample(x       = seq(prod(dim(myMatrix))), 
                  size    = as.integer(0.4*prod(dim(myMatrix))), 
                  replace = FALSE)
myMatrix[naIndex] <- NA

# Build kernel
radius <- 3

# Apply filters
meanfilterExample     <- meanFilter(X = myMatrix, radius = radius)
quantilefilterExample <- quantileFilter(X = myMatrix, radius = radius, probs = 0.1)
medianfilterExample   <- medianFilter(X = myMatrix, radius = radius)
```

Now, we will plot both data (original and filtered) using `image` function, as shows in figures 1 and 2.

**Original and Filtered**

```{r, message=FALSE, fig.height=5, fig.width=7.5, fig.cap = "Figure 2: Basic filters comparison", results='hide', fig.pos="h", echo=FALSE}
# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Add some noise (NA) to the image (matrix)
set.seed(7)
naIndex <- sample(x = seq(prod(dim(myMatrix))), size = as.integer(0.4*prod(dim(myMatrix))))
myMatrix[naIndex] <- NA

# Apply filters
meanfilterExample   <- meanFilter(X = myMatrix, radius = 3)
medianfilterExample <- medianFilter(X = myMatrix, radius = 3)

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(myMatrix, col = cols, axes = FALSE)
mtext(text = "Original with noise (NA)", side = 3, line = -1.5, font = 2, adj = 0.99)

# meanfilterExample[meanfilterExample < 0] <- 0
image(meanfilterExample, col = cols, axes = FALSE)
mtext(text = "Mean filter", side = 3, line = -1.5, font = 2, adj = 0.99)

# medianfilterExample[medianfilterExample < 0] <- 0
image(medianfilterExample, col = cols, axes = FALSE)
mtext(text = "2D median filter", side = 3, line = -1.5, font = 2, adj = 0.99)
```

## Kernel application

In the field of image processing, one of the tools most commonly used are the **convolutions**, which consist of operations between two arrays: The array of image data (as a big matrix) and kernels (as small matrices) which weighs each pixel values by the values of its corresponding neighborhood. Different kernels produce different effects, for instance: blur, shifted images (right, left, up or down), sharpening, etc. The users must be cautious with the size of the kernel because the larger the radius, the more pixels remain unanalyzed at the edges.

Besides, every function of **imagine** allows the recursive running of a filter by the using of `times` argument.

```{r, eval=FALSE}
medianFilter(X = wbImage, radius = 5, times = 50)
```


```{r, message=FALSE, fig.height=5, fig.width=7.5, fig.cap = "Figure 3: Filters with several time settings", results='hide', fig.pos="h", echo=FALSE}

times <- c(1, 5, 15)

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Apply filters
median_times1 <- medianFilter(X = wbImage, radius = 5, times = times[1])
median_times2 <- medianFilter(X = wbImage, radius = 5, times = times[2])
median_times3 <- medianFilter(X = wbImage, radius = 5, times = times[3])

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(median_times1, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[1]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times2, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[2]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times3, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[3]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)
```

## Filters based on published articles

Since its version 2.1.0, `imagine` includes two functions that implement the algorithms of two papers related to the calculation of oceanographic gradients: 

* `contextualMF`: Based on pseudocode provided in Belkin, I. M., & O'Reilly, J. E. (2009). An algorithm for oceanic front detection in chlorophyll and SST satellite imagery. Journal of Marine Systems, 78(3), 319-326. <https://doi.org/10.1016/j.jmarsys.2008.11.018>

* `agenbagFilters`: Based on Agenbag, J. J., Richardson, A. J., Demarcq, H., Freon, P., Weeks, S., & Shillington, F. A. (2003). Estimating environmental preferences of South African pelagic fish species using catch size- and remote sensing data. Progress in Oceanography, 59(2-3), 275-300. <https://doi.org/10.1016/j.pocean.2003.07.004>

Although both functions are available and can be executed directly from **imagine**, it is recommended to use them through the [**grec**](https://cran.r-project.org/package=grec) package.

---
title: "Introduction to imagine package"
vignette: >
  %\VignetteIndexEntry{Introduction to imagine package}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
knitr:
  opts_chunk:
    collapse: true
    comment: '#>'
---

```{r}
#| label: setup
library(imagine)
```

# IMAGing engINE: Tools for Application of Image Filters to Data Matrices

The **imagine** package streamlines the application of image-filtering algorithms to numeric data matrices. It implements efficient median-filtering and 2D convolution routines in C++ via Rcpp, enabling fast processing of large datasets.


## Installation

For installing **imagine**, as follows:

```{r}
#| eval: false
install.packages("imagine")
```

## Engines

The **imagine** package relies on C++-based algorithms, referred to as engines, implemented using the Rcpp and RcppArmadillo packages to accelerate filtering operations. These engines provide substantial performance improvements when working with large matrices.

As of version 2.1.0, the following engines are available:

* Engine 1: Standard 2D convolution (kernel applied to each cell neighborhood and summed).
* Engine 2: Same as Engine 1, but returns a quantile defined by the `probs` argument.
* Engine 3: Computes the mean over a square neighborhood defined by `radius` argument.
* Engine 4: Same as Engine 3, but returns a quantile defined by `probs` argument.
* Engine 5: Contextual Median Filter described in [Belkin and O'Reilly (2009)](http://dx.doi.org/10.1016/j.jmarsys.2008.11.018).
* Engines 6 and 7: Gradient filters described in [Agenbag et al. (2003)](https://doi.org/10.1016/j.pocean.2003.07.004).

Since version 2.0.0, the radius argument can accept two values, allowing rectangular (non-square) neighborhoods.

## Main functions

The package provides five main functions and two wrappers.

### Convolution functions
```{r}
#| eval: false

# Build kernels
# Kernel 1: For bottom edge recognition
kernel1 <- matrix(
  data = c(-1, -2, -1, 
            0,  0,  0,
            1,  2,  1), 
  nrow = 3
)

# Kernel 2: Diagonal weighting
kernel2 <- matrix(
  data = c(-2, 0, 0,
            0, 1, 0,
            0, 0, 2), 
  nrow = 3
)

# Apply filters
convolutionExample  <- convolution2D(X = wbImage, kernel = kernel1)
convQuantileExample <- convolutionQuantile(X = wbImage, kernel = kernel2, probs = 0.1)
```

To compare results, we plot the filtered outputs using the `image` function (@fig-2d-quantile-conv).

**Original vs filtered outputs**

```{r}
#| message: false
#| label: fig-2d-quantile-conv
#| fig-width: 5.33
#| fig-height: 8.5
#| fig-cap: "2D vs 2D quantile convolutions"
#| fig-pos: "h"
#| results: hide
#| echo: false

# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Build kernels
# Kernel 1: For bottom edge recognition
kernel1 <- matrix(
  data = c(-1, -2, -1,
            0,  0,  0,
            1,  2,  1), 
  nrow = 3
)

# Kernel 2: Diagonal weighting 
kernel2 <- matrix(
  data = c(-2, 0, 0,
            0, 1, 0,
            0, 0, 2), 
  nrow = 3
)

# Apply filters
convolutionExample  <- convolution2D(X = myMatrix, kernel = kernel1)
convQuantileExample <- convolutionQuantile(
  X = myMatrix, 
  kernel = kernel2, 
  probs = 0.8
)

# Make plots
par(
  mar = c(0, 0.5, 0, 0.5), 
  oma = c(0, 0, 2, 0), 
  mfrow = c(3, 1)
)

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(convolutionExample, col = cols, axes = FALSE)
mtext(
  text = "2D convolution", 
  side = 3, 
  line = -1.5, 
  col = "white", 
  font = 2, 
  adj = 0.99
)

image(convQuantileExample, col = cols, axes = FALSE)
mtext(
  text = "2D quantile convolution", 
  side = 3, 
  line = -1.5, 
  col = "black", 
  font = 2, 
  adj = 0.99
)
```


### Median-filter asociated functions
```{r}
#| eval: false

# Add some noise (NA) to the image (matrix)
set.seed(7)
naIndex <- sample(
  x = seq(prod(dim(myMatrix))), 
  size = as.integer(0.4*prod(dim(myMatrix))), 
  replace = FALSE
)
myMatrix[naIndex] <- NA

# Build kernel
radius <- 3

# Apply filters
meanfilterExample     <- meanFilter(X = myMatrix, radius = radius)
quantilefilterExample <- quantileFilter(X = myMatrix, radius = radius, probs = 0.1)
medianfilterExample   <- medianFilter(X = myMatrix, radius = radius)
```

We now compare the original and filtered matrices (@fig-basic-filter-comparison).

**Original and Filtered**

```{r}
#| message: false
#| label: fig-basic-filter-comparison
#| fig-width: 7.5
#| fig-height: 5
#| fig-cap: "Basic filters comparison"
#| fig-pos: "h"
#| results: hide
#| echo: false

# Defining a copy of wbImage
myMatrix <- wbImage

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Add some noise (NA) to the image (matrix)
set.seed(7)
naIndex <- sample(
  x = seq(prod(dim(myMatrix))), 
  size = as.integer(0.4*prod(dim(myMatrix)))
)
myMatrix[naIndex] <- NA

# Apply filters
meanfilterExample   <- meanFilter(X = myMatrix, radius = 3)
medianfilterExample <- medianFilter(X = myMatrix, radius = 3)

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(myMatrix, col = cols, axes = FALSE)
mtext(
  text = "Original with noise (NA)", 
  side = 3, 
  line = -1.5, 
  font = 2, 
  adj = 0.99
)

# meanfilterExample[meanfilterExample < 0] <- 0
image(meanfilterExample, col = cols, axes = FALSE)
mtext(
  text = "Mean filter", 
  side = 3, 
  line = -1.5, 
  font = 2, 
  adj = 0.99
)

# medianfilterExample[medianfilterExample < 0] <- 0
image(medianfilterExample, col = cols, axes = FALSE)
mtext(
  text = "2D median filter", 
  side = 3, 
  line = -1.5, 
  font = 2, 
  adj = 0.99
)
```

## Kernel application

In image processing, convolution is one of the most widely used operations. It consists of combining two arrays: the image (a matrix) and a kernel (a smaller matrix), which weights each pixel based on its neighborhood. Different kernels produce different effects, such as smoothing, shifting, or edge detection. Users should be cautious with kernel size, as larger neighborhoods reduce the number of valid pixels near the edges. All functions in **imagine** support recursive application through the times argument (@fig-recursive-filtering).

```{r}
#| eval: false

medianFilter(X = wbImage, radius = 5, times = 50)
```


```{r}
#| message: false
#| label: fig-recursive-filtering
#| fig-width: 7.5
#| fig-height: 5
#| fig-cap: "Filters with several time settings"
#| fig-pos: "h"
#| results: hide
#| echo: false

times <- c(1, 5, 15)

# Defining color palette
cols <- gray.colors(n = 1e3, start = 0, end = 1)

# Apply filters
median_times1 <- medianFilter(X = wbImage, radius = 5, times = times[1])
median_times2 <- medianFilter(X = wbImage, radius = 5, times = times[2])
median_times3 <- medianFilter(X = wbImage, radius = 5, times = times[3])

# Make plots
par(mar = rep(0, 4), oma = rep(0.5, 4), mfrow = c(2, 2))

image(wbImage, col = cols, axes = FALSE)
mtext(text = "Original", side = 3, line = -1.5, font = 2, adj = 0.99)

image(median_times1, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[1]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times2, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[2]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)

image(median_times3, col = cols, axes = FALSE)
mtext(text = paste("2D median filter\ntimes =", times[3]), 
      side = 3, line = -2.5, font = 2, adj = 0.99)
```

## Filters based on published articles

Since version 2.1.0, **imagine** includes implementations of algorithms from the literature for detecting oceanographic gradients:

* contextualMF: Based on [Belkin and O'Reilly (2009)](https://doi.org/10.1016/j.jmarsys.2008.11.018).
* agenbagFilters: Based on [Agenbag et al. (2003)](https://doi.org/10.1016/j.pocean.2003.07.004).

Although both functions are available directly in **imagine**, their use is recommended through the [grec](https://cran.r-project.org/package=grec) package for more specialized workflows.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imagine-main.R
\name{convolution2D}
\alias{convolution2D}
\alias{convolutionQuantile}
\alias{convolutionMedian}
\title{Make convolution calculations from numeric matrix}
\usage{
convolution2D(X, kernel, times = 1, normalize = FALSE, na_only = FALSE)

convolutionQuantile(
  X,
  kernel,
  probs,
  times = 1,
  normalize = FALSE,
  na_only = FALSE
)

convolutionMedian(X, kernel, times = 1, na_only = FALSE)
}
\arguments{
\item{X}{A numeric \code{matrix} object used for apply filters.}

\item{kernel}{A little matrix used as mask for each cell of \code{X}.}

\item{times}{How many times do you want to apply the filter?}

\item{normalize}{\code{logical} indicating if results will (or not) be
normalized. See details.}

\item{na_only}{\code{logical}, Do you want to apply the filter ONLY in cells
with NA?}

\item{probs}{\code{numeric} vector of probabilities with values in [0,1].}
}
\value{
\code{convolution2D} returns a \code{matrix} object with the same
dimensions of \code{X}.

\code{convolutionQuantile} uses the kernel but, for each cell, it
returns the position of quantile 'probs' (value between 0 and 1).

\code{convolutionMedian} is a wrapper of \code{convolutionQuantile}
with probs = 0.5.
}
\description{
This function takes a \code{matrix} object, and for each cell
multiplies its neighborhood by the \code{kernel}. Finally, it returns for
each cell the mean of the kernel-weighted sum.
}
\details{
Convolution is a mathematical operation that combines two arrays of numbers
to produce an array of the same structure. The output will consist of only
valid values, meaning those where both arrays have non-missing data.
Consequently, any missing values (NAs) in the input matrix will propagate
outwards to the extent of the convolution kernel.

Through normalization, the output of each convolution window is scaled by
dividing it by the sum of the absolute values of the kernel
(\code{sum(abs(as.numeric(kernel)))}, disabled by default).

\code{na_only} performs two actions at once: (1) it applies the filter only in
the positions where the original value is \code{NA} and (2) for the rest of
the cells, it is replaced with the value of the original matrix.
}
\examples{
# Generate example matrix
nRows <- 50
nCols <- 100

myMatrix <- matrix(runif(nRows*nCols, 0, 100), nrow = nRows, ncol = nCols)
kernel <- diag(3)

# Make convolution
myOutput1 <- convolution2D(myMatrix, kernel)
myOutput2 <- convolutionQuantile(myMatrix, kernel, probs = 0.7)

# Plot results
par(mfrow = c(2, 2))
image(myMatrix, zlim = c(0, 100))
image(myOutput1, zlim = c(0, 100))
image(myOutput2, zlim = c(0, 100))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imagine-main.R
\name{contextualMF}
\alias{contextualMF}
\title{Performs Contextual Median Filter}
\usage{
contextualMF(X)
}
\arguments{
\item{X}{A numeric \code{matrix} object used for apply filters.}
}
\value{
\code{contextualMF} returns a \code{matrix} object with the same
dimensions of \code{X}.
}
\description{
This function implements the Contextual Median Filter (CMF)
algorithm, which was first described by Belkin & O'Reilly (2009), following
the pseudocode provided in their paper.
}
\details{
Following the definition of CMF, since \strong{imagine} v.2.0.0, \code{times}
argument will not be available anymore.

\strong{imagine} offers the CMF algorithm but for the using to find out
oceanographic fronts, it is recommended to see and use the functions of the
\href{https://CRAN.R-project.org/package=grec}{\strong{grec}} package.
}
\examples{
data(wbImage)

cmfOut <- contextualMF(X = wbImage)

# image(cmfOut)
}
\references{
Belkin, I. M., & O'Reilly, J. E. (2009). An algorithm for oceanic
front detection in chlorophyll and SST satellite imagery. Journal of Marine
Systems, 78(3), 319-326 (\doi{http://dx.doi.org/10.1016/j.jmarsys.2008.11.018}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imagine-main.R
\name{agenbagFilters}
\alias{agenbagFilters}
\title{Performs algorithms from Agenbag et al. (2003)}
\usage{
agenbagFilters(X, algorithm = c(1, 2), ...)
}
\arguments{
\item{X}{A numeric \code{matrix} used as main input.}

\item{algorithm}{\code{integer} specifying the type of method that will be
used. See Details.}

\item{...}{Not used.}
}
\value{
\code{agenbagFilters} returns a \code{matrix} object with the same
dimensions of \code{X}.
}
\description{
This function performs two (gradient) calculation approaches for
SST, as outlined in the paper by Agenbag et al. (2003).
}
\details{
Section 2.2.4 of the paper by Agenbag et al. (2003) introduces the following
two methods:
\describe{
 \item{\strong{Method 1: }}{Based on the equation
 \deqn{Y_{i,j}=\sqrt{(X_{i+1,j}-X_{i-1,j})^2 +(X_{i,j+1}-X_{i,j-1})^2}}}
 where \eqn{Y_{i,j}} represents the output value for each \eqn{X_{i,j}} pixel value
 of a given \eqn{X} matrix.
 \item{\strong{Method 2: }}{the standard deviation in a 3x3 pixel area centered on
 position \eqn{(i,j)}.}
}

As outlined in the original study, this method conducts searches within a
1-pixel vicinity of each point. For method 1, it only returns a value for
points where none of the four involved values are NA. Conversely, for method
2, the standard deviation calculation is performed only for points where at
least 3 non-NA values are found in the 3x3 neighborhood.
}
\examples{
data(wbImage)

# Agenbag, method 1
agenbag1 <- agenbagFilters(X = wbImage, algorithm = 1)

# Agenbag, method 2
agenbag2 <- agenbagFilters(X = wbImage, algorithm = 2)

# Plotting results
par(mfrow = c(3, 1), mar = rep(0, 4))

# Original
image(wbImage, axes = FALSE, col = gray.colors(n = 1e3))

# Calculated
cols <- hcl.colors(n = 1e3, palette = "YlOrRd", rev = TRUE)
image(agenbag1, axes = FALSE, col = cols)
image(agenbag2, axes = FALSE, col = cols)
}
\references{
Agenbag, J.J., A.J. Richardson, H. Demarcq, P. Freon, S. Weeks,
and F.A. Shillington. "Estimating Environmental Preferences of South African
Pelagic Fish Species Using Catch Size- and Remote Sensing Data". Progress in
Oceanography 59, No 2-3 (October 2003): 275-300.
(\doi{https://doi.org/10.1016/j.pocean.2003.07.004}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imagine-main.R
\docType{data}
\name{wbImage}
\alias{wbImage}
\title{Data matrix to be used as example image.}
\format{
A \code{matrix} with dimensions 1280x720.
}
\usage{
wbImage
}
\description{
\code{matrix} object containing numeric data to plot a image.
The photo was taken by the author on 2016.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imagine-main.R
\name{meanFilter}
\alias{meanFilter}
\alias{quantileFilter}
\alias{medianFilter}
\title{Make a 2D filter calculations from numeric matrix}
\usage{
meanFilter(X, radius, times = 1, na_only = FALSE)

quantileFilter(X, radius, probs, times = 1, na_only = FALSE)

medianFilter(X, radius, times = 1, na_only = FALSE)
}
\arguments{
\item{X}{A numeric \code{matrix} object used for apply filters.}

\item{radius}{Size of squared or rectangular kernel to apply median. See
Details.}

\item{times}{How many times do you want to apply the filter?}

\item{na_only}{\code{logical}, Do you want to apply the filter ONLY in cells
with NA?}

\item{probs}{\code{numeric} vector of probabilities with values in [0,1].}
}
\value{
A \code{matrix} object with the same dimensions of \code{X}.

\code{quantileFilter} don't use a kernel but, for each cell, it
returns the position of quantile 'probs' (value between 0 and 1).

\code{medianFilter} is a wrapper of \code{quantileFilter} with
\code{probs = 0.5}.
}
\description{
This functions take a \code{matrix} object, and for each cell
calculate mean, median or certain quantile around a squared/rectangular
neighborhood.
}
\details{
\code{radius} must be defined as a 2-length numeric vector
specifying the number of rows and columns of the window which will be used to
make calculations. If the length of radius is 1, the window will be a square.

Functions use C++ algorithms for running some statistical calculations. The
mean is far obvious, however, there are several ways to perform quantiles.
\code{quantileFilter} function uses
\href{https://arma.sourceforge.net/docs.html#quantile}{arma::quantile}: a
RcppArmadillo function, which is equivalent to use R \link[stats]{quantile}
funtion with \code{type = 5}.

\code{medianFilter} is a wraper of \code{quantileFilter}, so the same
observations are applied to it.

\code{na_only} performs two actions at once: (1) it applies the filter only in
the positions where the original value is \code{NA} and (2) for the rest of
the cells, it is replaced with the value of the original matrix.
}
\examples{
# Generate example matrix
nRows <- 50
nCols <- 100

myMatrix <- matrix(runif(nRows*nCols, 0, 100), nrow = nRows, ncol = nCols)
radius <- 3

# Make convolution
myOutput1 <- meanFilter(X = myMatrix, radius = radius)
myOutput2 <- quantileFilter(X = myMatrix, radius = radius, probs = 0.1)
myOutput3 <- medianFilter(X = myMatrix, radius = radius)

# Plot results
par(mfrow = c(2, 2))
image(myMatrix, zlim = c(0, 100), title = "Original")
image(myOutput1, zlim = c(0, 100), title = "meanFilter")
image(myOutput2, zlim = c(0, 100), title = "quantileFilter")
image(myOutput3, zlim = c(0, 100), title = "medianFilter")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imagine-package.R
\docType{package}
\name{imagine-package}
\alias{imagine}
\alias{imagine-package}
\title{IMAGing engINE, Tools for application of image filters to data matrices}
\description{
Provides fast application of image filters to data matrices,
using R and C++ algorithms.
}
\details{
This package uses C++ algorithms called 'engines'. More details are
shown in the vignette.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/LuisLauM/imagine}
  \item Report bugs at \url{https://github.com/LuisLauM/imagine/issues}
}

}
\author{
\strong{Maintainer}: Wencheng Lau-Medrano \email{luis.laum@protonmail.com}

Authors:
\itemize{
  \item Wencheng Lau-Medrano \email{luis.laum@protonmail.com}
}

}
\keyword{internal}
